#!/usr/bin/python

from patches import parse_patches
import os
import sys
import string
import tty, termios

def parse_args(args):
    if len(args) == 2 and args[1] == '--bzr-usage':
        print '\n'
        return True
    elif len(args) == 2 and args[1] == '--bzr-help':
        print 'Shelve a patch, you can get it back later with unshelve.'
        return True
    elif len(args) == 1:
        pass
    else:
        raise Exception("Don't understand args %s" % args)

    return False

def unshelve():
    root = run_bzr('root')[0].strip()
    shelf = os.path.join(root, '.bzr-shelf')

    if not os.path.exists(shelf):
        raise Exception("No shelf found in '%s'" % shelf)

    patch = open(shelf, 'r').read()

    print >>sys.stderr, "Reapplying shelved patches"
    pipe = os.popen('patch -d %s -s -p1' % root, 'w')
    pipe.write(patch)
    pipe.flush()

    if pipe.close() is not None:
        raise Exception("Failed running patch!")

    os.remove(shelf)
    print 'Diff status is now:'
    os.system('bzr diff | diffstat')

    return True

class QuitException(Exception):
    pass

def shelve():
    patches = parse_patches(run_bzr('diff --hunks'))
    try:
        patches = HunkSelector(patches).select()
    except QuitException:
        return False

    if len(patches) == 0:
        print >>sys.stderr, 'Nothing to shelve'
        return True

    print patches
    return

    root = run_bzr('root')[0].strip()
    shelf = os.path.join(root, '.bzr-shelf')
    print >>sys.stderr, "Saving shelved patches to", shelf
    shelf = open(shelf, 'a')

    for patch in patches:
        shelf.write(str(patch))

    shelf.flush()
    os.fsync(shelf.fileno())
    shelf.close()

    print >>sys.stderr, "Reverting shelved patches"
    pipe = os.popen('patch -d %s -sR -p1' % root, 'w')
    for patch in patches:
       pipe.write(str(patch))
    pipe.flush()

    if pipe.close() is not None:
        raise Exception("Failed running patch!")

    print 'Diff status is now:'
    os.system('bzr diff | diffstat')

    return True

def run_bzr(args):
    if type(args) is str:
        args = [ args ]
    pipe = os.popen('bzr %s' % string.join(args, ' '), 'r')
    lines = pipe.readlines()
    if pipe.close() is not None:
        raise Exception("Failed running bzr")
    return lines


class HunkSelector:
    __usage = """Your options are:
    y - keep this change in your tree.
    n - shelve this change for the moment.
    i - invert the current selection.
    s - status, show what's selected to be shelved.
    q - quit the shelve command.\n"""

    def __init__(self, patches):
        self.patches = patches
        self.total_hunks = 0
        for patch in patches:
            self.total_hunks += len(patch.hunks)

    def select(self):
        j = 1
        for patch in self.patches:
            i = 0
            lasti = -1
            while i < len(patch.hunks):
                hunk = patch.hunks[i]
                if lasti != i:
                    print patch.get_header(), hunk
                lasti = i

                action = self.__ask_user(i + j)
                if action == 'keep':
                    hunk.selected = False
                elif action == 'shelve':
                    hunk.selected = True
                elif action == 'invert':
                    self.__invert_selection()
                    self.__show_status()
                    continue
                elif action == 'status':
                    self.__show_status()
                    continue
                elif action == 'quit':
                    return []

                i += 1

        for patch in self.patches:
            tmp = []
            for hunk in patch.hunks:
                if hunk.selected:
                    tmp.append(hunk)
            patch.hunks = tmp

        tmp = []
        for patch in self.patches:
            if len(patch.hunks):
                tmp.append(patch)
        self.patches = tmp

        return self.patches

    def __invert_selection(self):
        for patch in self.patches:
            for hunk in patch.hunks:
                if hunk.__dict__.has_key('selected'):
                    hunk.selected = not hunk.selected
                else:
                    hunk.selected = True

    def __show_status(self):
        print '\nStatus:'
        for patch in self.patches:
            print '  %s' % patch.oldname
            shelve = 0
            keep = 0
            unknown = 0
            for hunk in patch.hunks:
                if hunk.__dict__.has_key('selected'):
                    if hunk.selected:
                        shelve += 1
                    else:
                        keep += 1
                else:
                    unknown += 1

            print '    %d hunks to be shelved' % shelve
            print '    %d hunks to be kept' % keep
            print '    %d hunks undecided' % unknown
            print

    def __getchar(self):
        fd = sys.stdin.fileno()
        settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, settings)
        return ch

    def __ask_user(self, num):
        while True:
            sys.stdout.write('Keep this change? (%d of %d) [ynisq?] (n): ' %
                (num, self.total_hunks))

            response = self.__getchar()

            # default, which we see as newline, is 'n'
            if response in ['\n', '\r', '\r\n']:
                response = 'n'

            print response # because echo is off

            if response in ['y', 'Y']:
                return 'keep'
            elif response in ['n', 'N', '\n', '\r', '\r\n']:
                return 'shelve'
            elif response in ['i']:
                return 'invert'
            elif response in ['s']:
                return 'status'
            elif response in ['q', 'Q']:
                return 'quit'

            print self.__usage
